//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <Clipbrd.hpp>
#include <IniFiles.hpp>
#include "TMainForm.h"
#include "Common.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
static const AnsiString STR_TITLE = "Syncstamper";
static const AnsiString STR_ABOUT = STR_TITLE + " for Win32 Version 1.1.2\r\n\r\n"
	"Copyright (C) 2003 Moochos. All Rights Reserved.";
static const AnsiString STR_TIPS = "To cancel the process, press Escape key.\r\n"
	"To show Properties of Get data, double-click the status bar.";
static const AnsiString FILE_HELP = ExtractFilePath(Application->ExeName) + "Help\\index.html";
static const AnsiString FILE_INI = ExtractFilePath(Application->ExeName) + "Settings.ini";
static const AnsiString INI_SECTION = "Settings";
static const AnsiString INI_KEY = "Values";
static const AnsiString PARAM_DEFAULTPOSITION = "/dp";
//---------------------------------------------------------------------------
// struct TViewOptions
//---------------------------------------------------------------------------
__fastcall TViewOptions::TViewOptions() : MenuIcons(true), SwapPanels(false),
	StatusBar(true), Topmost(false), LastPosition(false)
{
	// do nothing
}
//---------------------------------------------------------------------------
__fastcall TViewOptions::~TViewOptions()
{
	// do nothing
}
//---------------------------------------------------------------------------
bool __fastcall TViewOptions::operator==(const TViewOptions& ASrc) const
{
	return MenuIcons == ASrc.MenuIcons &&
		SwapPanels == ASrc.SwapPanels &&
		StatusBar == ASrc.StatusBar &&
		Topmost == ASrc.Topmost &&
		LastPosition == ASrc.LastPosition;
}
//---------------------------------------------------------------------------
bool __fastcall TViewOptions::operator!=(const TViewOptions& ASrc) const
{
	return !(*this == ASrc);
}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
// class TTMainForm
//---------------------------------------------------------------------------
// published:
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	// check processing
	CanClose = !IsProcessing(true);
	if(!CanClose) return;

	// confirm quit
	if(FHolder->Count > 0 || FHistory->CanUndo() || FHistory->CanRedo()) {
		CanClose = Application->MessageBox("Quits the application.\r\n\r\nSure?",
			Application->Title.c_str(), MB_OKCANCEL | MB_ICONQUESTION) == IDOK;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	if(Key == VK_ESCAPE && Shift.Empty()) {
		if(IsProcessing())
			Cancel();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CheckBoxClick(TObject *Sender)
{
	TStampOptions stampOptions = StampOptions;
	if(Sender == CRTimeCheckBox) stampOptions.CRTime = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else if(Sender == LWTimeCheckBox) stampOptions.LWTime = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else if(Sender == LATimeCheckBox) stampOptions.LATime = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else if(Sender == SubdirsCheckBox) stampOptions.Subdirs = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else if(Sender == DirectoriesCheckBox) stampOptions.Directories = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else if(Sender == AnyAttrCheckBox) stampOptions.AnyAttr = dynamic_cast<TCheckBox*>(Sender)->Checked;
	else return;
	StampOptions = stampOptions;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StatusBarDblClick(TObject *Sender)
{
	if(!PropertiesMenuItem->Enabled) return;

	PropertiesMenuItem->Click();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PaintBoxPaint(TObject *Sender)
{
	int image;
	if(Sender == GetPaintBox) image = 0;
	else if(Sender == SetPaintBox) image = 1;
	else if(Sender == SyncPaintBox) image = 2;
	else return;

	TPaintBox* paintBox = dynamic_cast<TPaintBox*>(Sender);
	MenuImages->Draw(paintBox->Canvas, (paintBox->Width - MenuImages->Width) / 2,
		(paintBox->Height - MenuImages->Height) / 2, image, paintBox->Enabled);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MenuItemClick(TObject *Sender)
{
	if(IsProcessing(true)) return;

	if(Sender == NewWindowMenuItem || Sender == NewWindowPMenuItem) {
		SaveSettings();
		ShellExecute(Application->Handle, "open", Application->ExeName.c_str(),
			PARAM_DEFAULTPOSITION.c_str(), NULL, SW_SHOWNORMAL);
	}
	else if(Sender == UndoMenuItem || Sender == UndoPMenuItem) {
		Process(TStamper::ID_UNDO, NULL);
	}
	else if(Sender == RedoMenuItem || Sender == RedoPMenuItem) {
		Process(TStamper::ID_REDO, NULL);
	}
	else if(Sender == CopyErrorLogMenuItem || Sender == CopyErrorLogPMenuItem) {
		if(ErrorLog.IsEmpty()) {
			Application->MessageBox("No operation.\r\n\r\nNo error log.",
				Application->Title.c_str(), MB_ICONEXCLAMATION);
			return;
		}
		Clipboard()->AsText = ErrorLog;
	}
	else if(Sender == PropertiesMenuItem || Sender == PropertiesPMenuItem) {
		Process(TStamper::ID_PROPERTIES, NULL);
	}
	else if(Sender == ClearUndoRedoDataMenuItem) {
		if(Application->MessageBox("Clears Undo/Redo data.\r\n\r\nSure?",
			Application->Title.c_str(), MB_OKCANCEL | MB_ICONQUESTION) != IDOK) return;
		FHistory->Clear();
	}
	else if(Sender == ClearGetDataMenuItem) {
		if(Application->MessageBox("Clears Get data.\r\n\r\nSure?",
			Application->Title.c_str(), MB_OKCANCEL | MB_ICONQUESTION) != IDOK) return;
		FHolder->Clear();
		LowerPanel(NULL);
	}
	else if(Sender == MenuIconsMenuItem || Sender == SwapPanelsMenuItem || 
		Sender == StatusBarMenuItem || Sender == TopmostMenuItem || 
		Sender == LastPositionMenuItem) {
		TViewOptions viewOptions = ViewOptions;
		if(Sender == MenuIconsMenuItem) viewOptions.MenuIcons = !dynamic_cast<TMenuItem*>(Sender)->Checked;
		else if(Sender == SwapPanelsMenuItem) viewOptions.SwapPanels = !dynamic_cast<TMenuItem*>(Sender)->Checked;
		else if(Sender == StatusBarMenuItem) viewOptions.StatusBar = !dynamic_cast<TMenuItem*>(Sender)->Checked;
		else if(Sender == TopmostMenuItem) viewOptions.Topmost = !dynamic_cast<TMenuItem*>(Sender)->Checked;
		else if(Sender == LastPositionMenuItem) viewOptions.LastPosition = !dynamic_cast<TMenuItem*>(Sender)->Checked;
		ViewOptions = viewOptions;
	}
	else if(Sender == DefaultMenuItem) {
		if(Application->MessageBox("Applys default settings.\r\n\r\n"
			"Get and Undo/Redo data will not be cleared.\r\n\r\nSure?",
			Application->Title.c_str(), MB_OKCANCEL | MB_ICONQUESTION) != IDOK) return;
		DefaultSettings();
	}
	else if(Sender == ExitMenuItem) {
		Close();
	}
	else if(Sender == Help2MenuItem) {
		// confirm open
		if(Application->MessageBox("Opens a help file in HTML format.\r\n\r\nSure?",
			Application->Title.c_str(), MB_OKCANCEL | MB_ICONQUESTION) != IDOK) return;

		// open help file
		if(reinterpret_cast<unsigned int>(ShellExecute(Application->Handle, NULL,
			FILE_HELP.c_str(), NULL, NULL, SW_SHOWNORMAL)) <= 32) {
			Application->MessageBox(("Failed to open the file: " + FILE_HELP).c_str(),
				Application->Title.c_str(), MB_ICONEXCLAMATION);
		}
	}
	else if(Sender == TipsMenuItem) {
		Application->MessageBox(STR_TIPS.c_str(), "Tips", MB_ICONINFORMATION);
	}
	else if(Sender == AboutMenuItem) {
		Application->MessageBox(STR_ABOUT.c_str(), "About", MB_ICONINFORMATION);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::HolderChange(TObject *Sender)
{
	PropertiesMenuItem->Enabled = FHolder->Count > 0;
	ClearGetDataMenuItem->Enabled = PropertiesMenuItem->Enabled;
	PropertiesPMenuItem->Enabled = PropertiesMenuItem->Enabled;

	EnableControls(GetPanel, true);
	EnableControls(SetPanel, FHolder->Count > 0);
	EnableControls(SyncPanel, SetPanel->Enabled);

	GetPaintBox->Repaint();
	SetPaintBox->Repaint();
	SyncPaintBox->Repaint();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::HistoryChange(TObject *Sender)
{
	UndoMenuItem->Enabled = FHistory->CanUndo();
	RedoMenuItem->Enabled = FHistory->CanRedo();
	ClearUndoRedoDataMenuItem->Enabled = UndoMenuItem->Enabled || RedoMenuItem->Enabled;
	UndoPMenuItem->Enabled = UndoMenuItem->Enabled;
	RedoPMenuItem->Enabled = RedoMenuItem->Enabled;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ErrorLogChange(TObject *Sender)
{
	CopyErrorLogMenuItem->Enabled = !FErrorLog.IsEmpty();
	CopyErrorLogPMenuItem->Enabled = CopyErrorLogMenuItem->Enabled;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StampOptionsChange(TObject *Sender)
{
	CRTimeCheckBox->Checked = FStampOptions.CRTime;
	LWTimeCheckBox->Checked = FStampOptions.LWTime;
	LATimeCheckBox->Checked = FStampOptions.LATime;
	SubdirsCheckBox->Checked = FStampOptions.Subdirs;
	DirectoriesCheckBox->Checked = FStampOptions.Directories;
	AnyAttrCheckBox->Checked = FStampOptions.AnyAttr;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ViewOptionsChange(TObject *Sender)
{
	MenuIconsMenuItem->Checked = FViewOptions.MenuIcons;
	SwapPanelsMenuItem->Checked = FViewOptions.SwapPanels;
	StatusBarMenuItem->Checked = FViewOptions.StatusBar;
	TopmostMenuItem->Checked = FViewOptions.Topmost;
	LastPositionMenuItem->Checked = FViewOptions.LastPosition;

	FormMainMenu->Images = FViewOptions.MenuIcons ? MenuImages : NULL;
	FormPopupMenu->Images = FormMainMenu->Images;
	if(FViewOptions.SwapPanels != (LeftPanel->Left > RightPanel->Left)) {
		int temp = LeftPanel->Left;
		LeftPanel->Left = RightPanel->Left;
		RightPanel->Left = temp;
	}
	StatusBar->Visible = FViewOptions.StatusBar;
	SetWindowPos(Handle, FViewOptions.Topmost ? HWND_TOPMOST : HWND_NOTOPMOST,
		0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ThreadChange(TObject* ASender)
{
	static TWinControl* activeControl = ActiveControl;

	if(FThread == NULL) {
		Screen->Cursor = crDefault;
		ClientPanel->Enabled = true;
		DragAcceptFiles(Handle, TRUE);
		if(activeControl != NULL)
			activeControl->SetFocus();
	}
	else {
		activeControl = ActiveControl;
		Screen->Cursor = crHourGlass;
		ClientPanel->Enabled = false;
		DragAcceptFiles(Handle, FALSE);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StatusChange(TObject* ASender)
{
	StatusBar->SimpleText = (FStatusCount <= 0 ? AnsiString("") :
		(AnsiString(FStatusCount) + ": ")) + FStatusText;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::StamperTerminate(TObject* ASender)
{
	// process results
	TStamper* stamper = static_cast<TStamper*>(ASender);
	ErrorLog = stamper->ErrorLog->Text;

	// show results
	ShowStatus(StatusCount, TStamper::IdToName(stamper->Id) + " - " +
		(stamper->Canceled ? "Canceled " : "Done ") +
		FormatInt(stamper->Success) + "/" + FormatInt(stamper->Total) + " item(s)");

	Thread = NULL;
}
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetErrorLog(AnsiString AErrorLog)
{
	FErrorLog = AErrorLog;
	ErrorLogChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetStampOptions(TStampOptions AStampOptions)
{
	if(AStampOptions == FStampOptions) return;

	FStampOptions = AStampOptions;
	StampOptionsChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetViewOptions(TViewOptions AViewOptions)
{
	if(AViewOptions == FViewOptions) return;

	FViewOptions = AViewOptions;
	ViewOptionsChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetThread(TThread* AThread)
{
	if(AThread == FThread) return;

	FThread = AThread;
	ThreadChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ShowStatus(int AStatusCount, AnsiString AStatusText)
{
	FStatusCount = AStatusCount;
	FStatusText = AStatusText;
	StatusChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FilesDropped(TStrings* AFilePaths, TPoint AScreenPoint)
{
	if(IsProcessing(true)) return;

	// get id by dropped point
	TStamper::ID id;
	TPanel* panel;
	TPoint point;
	for(int i = 0; i < 4; i++) {
		switch(i) {
			case 0: id = TStamper::ID_GET; panel = GetPanel; break;
			case 1: id = TStamper::ID_SET; panel = SetPanel; break;
			case 2: id = TStamper::ID_SYNC; panel = SyncPanel; break;
			case 3: id = TStamper::ID_PROPERTIES; continue;
		}
		point = panel->ScreenToClient(AScreenPoint);
		if(point.x < 0 || point.x >= panel->Width || point.y < 0 ||
			point.y >= panel->Height || !panel->Enabled) continue;
		LowerPanel(panel);
		break;
	}

	// process file pahts
	Process(id, AFilePaths);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Process(TStamper::ID AId, TStrings* AFilePaths)
{
	if(IsProcessing(true)) return;

	// show status
	ShowStatus(StatusCount + 1, TStamper::IdToName(AId) + " - Press [Esc] to cancel");

	// start thread
	Thread = new TStamper(AId, AFilePaths, FHolder, FHistory, StampOptions);
	Thread->OnTerminate = StamperTerminate;
	Thread->Resume();
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::IsProcessing(bool AShowMessage) const
{
	if(FThread == NULL) return false;

	if(AShowMessage) {
		Application->MessageBox("Can't do it while processing.",
			Application->Title.c_str(), MB_ICONEXCLAMATION);
	}

	return true;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Cancel()
{
	if(!IsProcessing()) return;

	ShowStatus(StatusCount, "Canceling...");
	Thread->Terminate();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LowerPanel(TPanel* APanel)
{
	TPanel* panel;
	for(int i = 0; i < 3; i++) {
		switch(i) {
			case 0: panel = GetPanel; break;
			case 1: panel = SetPanel; break;
			case 2: panel = SyncPanel; break;
		}
		panel->BevelOuter = panel == APanel ? bvLowered : bvRaised;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EnableControls(TControl* AControl, bool AEnabled) const
{
	AControl->Enabled = AEnabled;

	if(!AControl->InheritsFrom(__classid(TWinControl))) return;
	TWinControl* control = dynamic_cast<TWinControl*>(AControl);
	for(int i = 0; i < control->ControlCount; i++)
		EnableControls(control->Controls[i], AEnabled);
}
//---------------------------------------------------------------------------
int __fastcall TMainForm::FindParam(AnsiString AParam, bool ACaseSensitive) const
{
	typedef int (__stdcall* TCompare)(LPCSTR, LPCSTR);
	TCompare compare = ACaseSensitive ? lstrcmp : lstrcmpi;
	for(int i = 0; i < ParamCount(); i++) {
		if(compare(ParamStr(i + 1).c_str(), AParam.c_str()) == 0)
			return i + 1;
	}

	return -1;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DefaultSettings()
{
	StampOptions = TStampOptions();
	ViewOptions = TViewOptions();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadSettings()
{
	TMemIniFile* ini = NULL;
	TStringList* list = new TStringList();
	int index = 0;
	try {
		ini = new TMemIniFile(FILE_INI);
		list->CommaText = ini->ReadString(INI_SECTION, INI_KEY, "");

		TStampOptions stampOptions;
		stampOptions.Subdirs = StrToBool(list->Strings[index++]);
		stampOptions.Directories = StrToBool(list->Strings[index++]);
		stampOptions.AnyAttr = StrToBool(list->Strings[index++]);
		stampOptions.CRTime = StrToBool(list->Strings[index++]);
		stampOptions.LWTime = StrToBool(list->Strings[index++]);
		stampOptions.LATime = StrToBool(list->Strings[index++]);
		StampOptions = stampOptions;

		TViewOptions viewOptions;
		viewOptions.MenuIcons = StrToBool(list->Strings[index++]);
		viewOptions.SwapPanels = StrToBool(list->Strings[index++]);
		viewOptions.StatusBar = StrToBool(list->Strings[index++]);
		viewOptions.Topmost = StrToBool(list->Strings[index++]);
		viewOptions.LastPosition = StrToBool(list->Strings[index++]);
		ViewOptions = viewOptions;

		int left = StrToInt(list->Strings[index++]);
		int top = StrToInt(list->Strings[index++]);
		if(ViewOptions.LastPosition && FindParam(PARAM_DEFAULTPOSITION, false) < 0) {
			Left = left;
			Top = top;
		}
	}
	catch(...) {
		DefaultSettings();
	}
	delete list;
	delete ini;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SaveSettings()
{
	TMemIniFile* ini = NULL;
	TStringList* list = new TStringList();
	try {
		ini = new TMemIniFile(FILE_INI);

		TStampOptions stampOptions = StampOptions;
		list->Add(BoolToStr(stampOptions.Subdirs));
		list->Add(BoolToStr(stampOptions.Directories));
		list->Add(BoolToStr(stampOptions.AnyAttr));
		list->Add(BoolToStr(stampOptions.CRTime));
		list->Add(BoolToStr(stampOptions.LWTime));
		list->Add(BoolToStr(stampOptions.LATime));

		TViewOptions viewOptions = ViewOptions;
		list->Add(BoolToStr(viewOptions.MenuIcons));
		list->Add(BoolToStr(viewOptions.SwapPanels));
		list->Add(BoolToStr(viewOptions.StatusBar));
		list->Add(BoolToStr(viewOptions.Topmost));
		list->Add(BoolToStr(viewOptions.LastPosition));

		list->Add(Left);
		list->Add(Top);
		list->Add("");

		ini->WriteString(INI_SECTION, INI_KEY, list->CommaText);
		ini->UpdateFile();
	}
	catch(...) {}
	delete list;
	delete ini;
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TMainForm::WMDropFiles(TWMDropFiles& AMessage)
{
	SetForegroundWindow(Application->Handle);

	// handle dropped files
	int fileCount = DragQueryFile(reinterpret_cast<HDROP>(AMessage.Drop), 0xFFFFFFFF, NULL, 0);
	if(fileCount > 0) {
		TStringList* filePaths = new TStringList();
		char filePath[MAX_PATH];
		for(int i = 0; i < fileCount; i++) {
			if(DragQueryFile(reinterpret_cast<HDROP>(AMessage.Drop), i, filePath, MAX_PATH) > 0)
				filePaths->Add(filePath);
		}
		TPoint dropPoint;
		DragQueryPoint(reinterpret_cast<HDROP>(AMessage.Drop), &dropPoint);
		FilesDropped(filePaths, ClientToScreen(dropPoint));
		delete filePaths;
	}
	DragFinish(reinterpret_cast<HDROP>(AMessage.Drop));
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner) : TForm(Owner), FHolder(NULL),
	FHistory(NULL), FThread(NULL), FStatusCount(0), FStatusText("Ready")
{
	// init
	Application->Title = STR_TITLE;
	Caption = STR_TITLE;
	StatusBar->Height = abs(Font->Height) + 12;
	ClientPanel->AutoSize = true;
	AutoSize = true;
	FHolder = new TStampHolder();
	FHolder->OnChange = HolderChange;
	FHistory = new TStampHistory();
	FHistory->OnChange = HistoryChange;

	// update
	HolderChange(this);
	HistoryChange(this);
	ErrorLogChange(this);
	StampOptionsChange(this);
	ViewOptionsChange(this);
	ThreadChange(this);
	StatusChange(this);
	LowerPanel(NULL);

	LoadSettings();
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
	SaveSettings();

	// uninit
	delete FHistory;
	delete FHolder;
}
//---------------------------------------------------------------------------

