//---------------------------------------------------------------------------
#ifndef TStampH
#define TStampH
//---------------------------------------------------------------------------
#include <values.h>  // contains BITSPERBYTE
#include <vector>
#include <Classes.hpp>
//---------------------------------------------------------------------------
struct TStamp
{
	AnsiString FilePath;
	FILETIME CRTime, LWTime, LATime;
	__fastcall TStamp();
	__fastcall virtual ~TStamp();
	bool __fastcall operator==(const TStamp& ASrc) const;
	bool __fastcall operator!=(const TStamp& ASrc) const;
};
//---------------------------------------------------------------------------
typedef std::vector<TStamp> TStamps;
//---------------------------------------------------------------------------
struct TStampOptions
{
	bool Subdirs;
	bool Directories;
	bool AnyAttr;
	bool CRTime, LWTime, LATime;
	__fastcall TStampOptions();
	__fastcall virtual ~TStampOptions();
	bool __fastcall operator==(const TStampOptions& ASrc) const;
	bool __fastcall operator!=(const TStampOptions& ASrc) const;
};
//---------------------------------------------------------------------------
struct TStampUnit
{
	TStamps Stamps;
	TStampOptions Options;
	__fastcall TStampUnit();
	__fastcall virtual ~TStampUnit();
};
//---------------------------------------------------------------------------
typedef std::vector<TStampUnit> TStampUnits;
//---------------------------------------------------------------------------
class TStampHolder : public TObject
{
private:
	TStamps FStamps;
	TStamp FAverage;
	int FDirCount;
	TNotifyEvent FOnChange;
	const TStamps& __fastcall GetStamps() const;
	const TStamp& __fastcall GetAverage() const;
	int __fastcall GetCount() const;
	int __fastcall GetFileCount() const;
protected:
	void __fastcall DoChange();
public:
	__fastcall TStampHolder();
	__fastcall virtual ~TStampHolder();
	void __fastcall Set(const TStamps& AStamps, const TStamp& AAverage, int ADirCount);
	void __fastcall Clear();
	__property const TStamps& Stamps = {read=GetStamps};
	__property const TStamp& Average = {read=GetAverage};
	__property int Count = {read=GetCount};
	__property int FileCount = {read=GetFileCount};
	__property int DirCount = {read=FDirCount};
	__property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
};
//---------------------------------------------------------------------------
class TStampHistory : public TObject
{
private:
	TStampUnits FValues;
	unsigned int FIndex;
	TNotifyEvent FOnChange;
	const TStampUnit& __fastcall GetUndoValue() const;
	const TStampUnit& __fastcall GetRedoValue() const;
protected:
	void __fastcall DoChange();
public:
	__fastcall TStampHistory();
	__fastcall virtual ~TStampHistory();
	bool __fastcall CanUndo() const;
	bool __fastcall CanRedo() const;
	void __fastcall Add(const TStampUnit& ASet);
	void __fastcall Undo(const TStampUnit& ASet);
	void __fastcall Redo(const TStampUnit& ASet);
	void __fastcall RemoveUndoValue();
	void __fastcall RemoveRedoValue();
	void __fastcall Clear();
	__property const TStampUnit& UndoValue = {read=GetUndoValue};
	__property const TStampUnit& RedoValue = {read=GetRedoValue};
	__property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
};
//---------------------------------------------------------------------------
static const int DWORD_BIT = BITSPERBYTE * sizeof(DWORD);
bool __fastcall IsValidFileTime(const FILETIME& AFileTime);
void __fastcall ClearFileTime(FILETIME& AFileTime);
DWORDLONG __fastcall FileTimeToDwordLong(const FILETIME& AFileTime);
FILETIME __fastcall DwordLongToFileTime(DWORDLONG ADwordLong);
//---------------------------------------------------------------------------
#endif