//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TStamper.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// class TStamper : public TStamperBase
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
bool __fastcall TStamper::Execute_CheckConditions()
{
	FTotal = FFilePaths == NULL ? 0 : FFilePaths->Count;

	// check time parapeters
	if((FId == ID_SET || FId == ID_SYNC) && !FOptions.CRTime && !FOptions.LWTime &&
		!FOptions.LATime) {
		ShowMessageBox("No operation.\r\n\r\nNo time parameter to process.",
			MB_ICONEXCLAMATION);
		return false;
	}

	return true;
}
//---------------------------------------------------------------------------
bool __fastcall TStamper::Execute_FilePaths()
{
	if(FFilePaths == NULL) return true;

	// prepare file paths
	PrepareFilePaths(FFilePaths, FOptions);
	if(Terminated) return false;
	FTotal = FFilePaths->Count;
	if(FTotal <= 0) {
		ShowMessageBox("No operation.\r\n\r\nNo item to process.", MB_ICONEXCLAMATION);
		return false;
	}

	return true;
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Get()
{
	// show properties
	if(FId == ID_PROPERTIES && FFilePaths == NULL) {
		int fileCount, dirCount;
		TStamp average;
		FParams->Clear();
		FParams->Add(&fileCount);
		FParams->Add(&dirCount);
		FParams->Add(&average);
		Synchronize(Execute_Get_GetProperties);
		ShowProperties(fileCount, dirCount, average);
		FTotal = FSuccess = fileCount + dirCount;
		return;
	}

	// declare
	TStamps stamps;
	TStamp average;
	int dirCount = 0;

	// prepare values
	stamps.resize(FFilePaths->Count);
	for(int i = 0; i < FFilePaths->Count; i++)
		stamps[i].FilePath = FFilePaths->Strings[i];
	FTotal = stamps.size();
	FSuccess = Get(FErrorLog, stamps, dirCount);
	if(Terminated) return;
	if(FSuccess <= 0) {
		if(FId == ID_GET)
			ShowMessageBox("No operation.\r\n\r\nFailed all items.\r\n\r\n"
				"Get data is not registered.", MB_ICONEXCLAMATION);
		else
			ShowMessageBox("No operation.\r\n\r\nFailed to get properties.",
				MB_ICONEXCLAMATION);
		return;
	}
	average = StampsToAverage(stamps);
	if(Terminated) return;

	// process accordingly
	if(FId == ID_GET) {
		FParams->Clear();
		FParams->Add(&stamps);
		FParams->Add(&average);
		FParams->Add(&dirCount);
		Synchronize(Execute_Get_Update);
	}
	else if(FId == ID_PROPERTIES) {
		ShowProperties(static_cast<int>(stamps.size()) - dirCount, dirCount, average);
	}
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Get_GetProperties()
{
	*static_cast<int*>(FParams->Items[0]) = FHolder->FileCount;
	*static_cast<int*>(FParams->Items[1]) = FHolder->DirCount;
	*static_cast<TStamp*>(FParams->Items[2]) = FHolder->Average;
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Get_Update()
{
	FHolder->Set(*static_cast<TStamps*>(FParams->Items[0]),
		*static_cast<TStamp*>(FParams->Items[1]),
		*static_cast<int*>(FParams->Items[2]));
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Set()
{
	// declare
	TStampUnit stampUnit, backup;

	// prepare stampUnit
	if(FFilePaths != NULL) FTotal = FFilePaths->Count;
	FParams->Clear();
	FParams->Add(&stampUnit);
	Synchronize(Execute_Set_GetStampUnit);
	if(static_cast<int>(stampUnit.Stamps.size()) <= 0) {
		ShowMessageBox("No operation.\r\n\r\nNo data to process.", MB_ICONEXCLAMATION);
		return;
	}
	else if(FId == ID_SYNC && static_cast<int>(stampUnit.Stamps.size()) != FFilePaths->Count) {
		AnsiString text = AnsiString("") +
			"No operation.\r\n\r\n"
			"Item count of source and target must be same.\r\n\r\n"
			"  Source: " + stampUnit.Stamps.size() + " item(s)\r\n"
			"  Target: " + FFilePaths->Count + " item(s)";
		ShowMessageBox(text, MB_ICONEXCLAMATION);
		return;
	}
	if(FId == ID_SET || FId == ID_SYNC) {
		for(int i = 0; i < FFilePaths->Count; i++)
			stampUnit.Stamps[i].FilePath = FFilePaths->Strings[i];
	}
	FTotal = stampUnit.Stamps.size();

	// prepare backup
	backup = stampUnit;
	int dirCount = 0;
	bool* valid = new bool[backup.Stamps.size()];
	for(unsigned int i = 0; i < backup.Stamps.size(); i++)
		valid[i] = true;
	int success = Get(FErrorLog, backup.Stamps, dirCount, valid);
	if(Terminated) {
		delete [] valid;
		return;
	}
	if(success < FTotal) {
		AnsiString text = AnsiString("") +
			"Failed to backup Undo/Redo data.\r\n\r\n"
			"If succeeded no item, Undo/Redo data is not registered.\r\n\r\n"
			"  Succeeded: " + success + "/" + FTotal + " item(s)\r\n\r\n"
			"Continue?";
		if(ShowMessageBox(text, MB_OKCANCEL | MB_ICONQUESTION) != IDOK) {
			Terminate();
			delete [] valid;
			return;
		}
	}

	// process files
	FSuccess = Set(FErrorLog, stampUnit.Stamps, stampUnit.Options, FId != ID_UNDO, valid);
	if(Terminated) {
		delete [] valid;
		return;
	}
	TrimStamps(backup.Stamps, valid);
	TrimStamps(stampUnit.Stamps, valid);
	delete [] valid;

	// register Undo/Redo data
	if(FSuccess <= 0) {
		ShowMessageBox("Succeeded no item.\r\n\r\nUndo/Redo data is not registered.",
			MB_ICONEXCLAMATION);
		backup.Stamps.clear();
	}
	FParams->Clear();
	FParams->Add(&backup);
	Synchronize(Execute_Set_SetToHistory);
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Set_GetStampUnit()
{
	TStampUnit* stampUnit = static_cast<TStampUnit*>(FParams->Items[0]);
	switch(FId) {
		case ID_SET:
		case ID_SYNC:      
			if(FId == ID_SET)
				stampUnit->Stamps.resize(FFilePaths->Count, FHolder->Average);
    		else
				stampUnit->Stamps = FHolder->Stamps;
			stampUnit->Options = FOptions;
			break;
		case ID_UNDO: *stampUnit = FHistory->UndoValue; break;
		case ID_REDO: *stampUnit = FHistory->RedoValue; break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute_Set_SetToHistory()
{
	TStampUnit* stampUnit = static_cast<TStampUnit*>(FParams->Items[0]);
	if(stampUnit->Stamps.size() <= 0) {
		switch(FId) {
			case ID_UNDO: FHistory->RemoveUndoValue(); break;
			case ID_REDO: FHistory->RemoveRedoValue(); break;
		}
		return;
	}
	switch(FId) {
		case ID_SET:
		case ID_SYNC: FHistory->Add(*stampUnit); break;
		case ID_UNDO: FHistory->Undo(*stampUnit); break;
		case ID_REDO: FHistory->Redo(*stampUnit); break;
	}
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TStamper::Execute()
{
	try {
		if(!Execute_CheckConditions() || Terminated) return;

		if(!Execute_FilePaths() || Terminated) return;

		switch(FId) {
			case ID_GET:
			case ID_PROPERTIES: Execute_Get(); break;
			case ID_SET:
			case ID_SYNC:
			case ID_UNDO:
			case ID_REDO: Execute_Set(); break;
		}
	}
	catch(...) {}
}
//---------------------------------------------------------------------------
void __fastcall TStamper::DoTerminate()
{
	FCanceled = Terminated;
	TStamperBase::DoTerminate();

	// uninit
	delete FParams;
	delete FErrorLog;
	delete FFilePaths;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TStamper::TStamper(ID AId, TStrings* AFilePaths, TStampHolder* AHolder,
	TStampHistory* AHistory, TStampOptions AOptions) : TStamperBase(),
	FTotal(0), FSuccess(0), FCanceled(false), FErrorLog(NULL), FId(AId),
	FFilePaths(NULL), FHolder(AHolder), FHistory(AHistory), FOptions(AOptions),
	FParams(NULL)
{
	// init
	if(AFilePaths != NULL) {
		FFilePaths = new TStringList();
		FFilePaths->Assign(AFilePaths);
	}
	FErrorLog = new TStringList();
	FParams = new TList();
}
//---------------------------------------------------------------------------
__fastcall TStamper::~TStamper()
{
	// do nothing
}
//---------------------------------------------------------------------------
AnsiString __fastcall TStamper::IdToName(ID AId)
{
	switch(AId) {
		case ID_GET: return "Get";
		case ID_SET: return "Set";
		case ID_SYNC: return "Sync";
		case ID_UNDO: return "Undo";
		case ID_REDO: return "Redo";
		case ID_PROPERTIES: return "Properties";
		default: return "(Unknown)";
	}
}
//---------------------------------------------------------------------------
