//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TStamperBase.h"
#include "Common.h"
#pragma package(smart_init)
using namespace std;
//---------------------------------------------------------------------------
// class TStamperBase : public TThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
AnsiString __fastcall TStamperBase::FilePathToDisplayName(const AnsiString& AFilePath) const
{
	AnsiString displayName = AFilePath;

	// convert file path to display name
	IMalloc* iMalloc;
	if(SHGetMalloc(&iMalloc) != NOERROR) return displayName;
	IShellFolder* iShellFolder;
	WCHAR buffer[MAX_PATH];
	ULONG eaten;
	ITEMIDLIST* itemIDList;
	char filePath[MAX_PATH];
	if(SHGetDesktopFolder(&iShellFolder) == NOERROR) {
		MultiByteToWideChar(CP_ACP, 0, AFilePath.c_str(), -1, buffer, MAX_PATH);
		if(iShellFolder->ParseDisplayName(NULL, NULL, buffer, &eaten, &itemIDList, NULL) == S_OK) {
			if(SHGetPathFromIDList(itemIDList, filePath)) {
				displayName = filePath;
			}
			iMalloc->Free(itemIDList);
		}
		iShellFolder->Release();
	}
	iMalloc->Release();

	return displayName;
}
//---------------------------------------------------------------------------
void __fastcall TStamperBase::AddSubdirs(TStrings* AFilePaths,
	const AnsiString& AFilePath, const TStampOptions& AOptions) const
{
	AnsiString filePath;
	WIN32_FIND_DATA data;
	HANDLE handle = FindFirstFile((AFilePath + "\\*").c_str(), &data);
	if(handle == INVALID_HANDLE_VALUE || handle == NULL) return;
	do {
		if(lstrcmpi(data.cFileName, ".") == 0 || lstrcmpi(data.cFileName, "..") == 0)
			continue;

		filePath = AFilePath + "\\" + data.cFileName;
		if((data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
			if(AOptions.Directories) AFilePaths->Add(filePath);
			if(AOptions.Subdirs) AddSubdirs(AFilePaths, filePath, AOptions);
		}
		else {
			AFilePaths->Add(filePath);
		}
	} while(FindNextFile(handle, &data) != FALSE && !Terminated);
	FindClose(handle);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TStamperBase::AttrsToStr(DWORD AAttrs) const
{
	const int count = 7;
	DWORD flags[count] = {
		FILE_ATTRIBUTE_ARCHIVE,
		FILE_ATTRIBUTE_DIRECTORY,
		FILE_ATTRIBUTE_HIDDEN,
		FILE_ATTRIBUTE_NORMAL,
		FILE_ATTRIBUTE_READONLY,
		FILE_ATTRIBUTE_SYSTEM,
		FILE_ATTRIBUTE_TEMPORARY};
	char initials[count] = {'A', 'D', 'H', 'N', 'R', 'S', 'T'};

	AnsiString str;
	for(int i = 0; i < count; i++)
		str += ((AAttrs & flags[i]) == flags[i] ? '+' : '-') + initials[i];

	return str;
}
//---------------------------------------------------------------------------
bool __fastcall TStamperBase::CalculateSum(DWORD* ASum, int ASize, const FILETIME& AFileTime) const
{
	if(!IsValidFileTime(AFileTime)) return false;

	CalculateSum(ASum, ASize, 0, AFileTime.dwLowDateTime);
	CalculateSum(ASum, ASize, 1, AFileTime.dwHighDateTime);

	return true;
}
//---------------------------------------------------------------------------
void __fastcall TStamperBase::CalculateSum(DWORD* ASum, int ASize, int APosition, DWORD AValue) const
{
	ASum[APosition] += AValue;
	if(ASum[APosition] < AValue && APosition < ASize - 1)
		CalculateSum(ASum, ASize, APosition + 1, 1);
}
//---------------------------------------------------------------------------
void __fastcall TStamperBase::CalculateAverage(FILETIME& AFileTime, DWORD* ASum, int ASize, int ATotal) const
{
	if(ATotal <= 0) {
		ClearFileTime(AFileTime);
		return;
	}

	int position;
	DWORDLONG n;
	for(int i = 0; i < ASize; i++) {
		position = ASize - 1 - i;
		n = ASum[position] + (i > 0 ? static_cast<DWORDLONG>(ASum[position + 1]) << DWORD_BIT : 0);
		switch(position) {
			case 0: AFileTime.dwLowDateTime = static_cast<DWORD>(n / ATotal); break;
			case 1: AFileTime.dwHighDateTime = static_cast<DWORD>(n / ATotal); break;
		}
		ASum[position] = static_cast<DWORD>(n % ATotal);
	}
}
//---------------------------------------------------------------------------
void __fastcall TStamperBase::ShowMessageBox_GetResult()
{
	FMessageBoxResult = Application->MessageBox(FMessageBoxText.c_str(),
		Application->Title.c_str(), FMessageBoxFlags);
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TStamperBase::PrepareFilePaths(TStrings* AFilePaths, TStampOptions AOptions) const
{
	TStringList* filePaths = new TStringList();
	filePaths->CaseSensitive = false;
	filePaths->Duplicates = dupIgnore;
	filePaths->Sorted = true;
	AnsiString filePath;
	for(int i = 0; i < AFilePaths->Count && !Terminated; i++) {
		filePath = FilePathToDisplayName(AFilePaths->Strings[i]);
		if(DirectoryExists(filePath)) {
			if(AOptions.Directories) filePaths->Add(filePath);
			if(AOptions.Subdirs) AddSubdirs(filePaths, filePath, AOptions);
		}
		else {
			filePaths->Add(filePath);
		}
	}
	AFilePaths->Assign(filePaths);
	delete filePaths;
}
//---------------------------------------------------------------------------
void __fastcall TStamperBase::TrimStamps(TStamps& AStamps, const bool* AValid) const
{
	const bool* valid = AValid;
	for(TStamps::iterator it = AStamps.begin(); it != AStamps.end() && !Terminated; it++, valid++) {
		if(*valid) continue;
		it = AStamps.erase(it);
		if(it == AStamps.end()) break;
		it--;
	}
}
//---------------------------------------------------------------------------
int __fastcall TStamperBase::Get(TStrings* AErrorLog, TStamps& AStamps,
	int& ADirCount) const
{
	if(AStamps.size() <= 0) return 0;

	bool* valid = new bool[AStamps.size()];
	for(unsigned int i = 0; i < AStamps.size(); i++)
		valid[i] = true;
	int ret = Get(AErrorLog, AStamps, ADirCount, valid);
	TrimStamps(AStamps, valid);
	delete [] valid;

	return ret;
}
//---------------------------------------------------------------------------
int __fastcall TStamperBase::Get(TStrings* AErrorLog, TStamps& AStamps,
	int& ADirCount, bool* AValid) const
{
	bool* valid = AValid;
	HANDLE handle;
	WIN32_FIND_DATA data;
	ADirCount = 0;
	int success = 0;
	for(TStamps::iterator it = AStamps.begin(); it != AStamps.end() && !Terminated; it++, valid++) {
		if(!*valid) continue;
		handle = FindFirstFile(it->FilePath.c_str(), &data);
		if(handle == INVALID_HANDLE_VALUE || handle == NULL) {
			AErrorLog->Add("failed to get data: " + it->FilePath);
			*valid = false;
			continue;
		}
		FindClose(handle);
		it->CRTime = data.ftCreationTime;
		it->LWTime = data.ftLastWriteTime;
		it->LATime = data.ftLastAccessTime;
		if((data.dwFileAttributes && FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
			ADirCount++;
		success++;
	}

	return success;
}
//---------------------------------------------------------------------------     
int __fastcall TStamperBase::Set(TStrings* AErrorLog, const TStamps& AStamps,
	TStampOptions AOptions, bool AForward, bool* AValid) const
{
	unsigned int index;
	const TStamp* stamp;
	bool* valid;
	DWORD attrs;
	bool attrsNormalized;
	HANDLE handle;
	int success = 0;
	for(unsigned int i = 0; i < AStamps.size() && !Terminated; i++) {
		index = AForward ? i : (AStamps.size() - 1 - i);
		stamp = &AStamps[index];
		valid = AValid + index;
		if(!*valid) continue;

		// normalize file attributes
		attrsNormalized = false;
		if(AOptions.AnyAttr) {
			attrs = GetFileAttributes(stamp->FilePath.c_str());
			if(attrs == 0xFFFFFFFF) {
				AErrorLog->Add("failed to get file attributes: " + stamp->FilePath);
			}
			else if(attrs != FILE_ATTRIBUTE_NORMAL && attrs != FILE_ATTRIBUTE_DIRECTORY &&
				SetFileAttributes(stamp->FilePath.c_str(), FILE_ATTRIBUTE_NORMAL) == FALSE) {
				AErrorLog->Add("failed to normalize file attributes: " + stamp->FilePath);
			}
			else {
				attrsNormalized = true;
			}
		}

		// set file times
		*valid = false;
		handle = CreateFile(stamp->FilePath.c_str(), GENERIC_WRITE,
			FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL |
			(DirectoryExists(stamp->FilePath) ? FILE_FLAG_BACKUP_SEMANTICS : 0), NULL);
		if(handle == INVALID_HANDLE_VALUE || handle == NULL) {
			AErrorLog->Add("failed write access: " + stamp->FilePath);
		}
		else {
			if(SetFileTime(handle,
				AOptions.CRTime ? &(stamp->CRTime) : NULL,
				AOptions.LATime ? &(stamp->LATime) : NULL,
				AOptions.LWTime ? &(stamp->LWTime) : NULL) == FALSE) {
				AErrorLog->Add("failed to set file times: " + stamp->FilePath);
			}
			else {
				*valid = true;
			}
			CloseHandle(handle);
		}

		// restore file attributes
		if(attrsNormalized) {
			if(SetFileAttributes(stamp->FilePath.c_str(), attrs) == FALSE) {
				AErrorLog->Add("failed to restore file attributes: " +
					stamp->FilePath + " (" + AttrsToStr(attrs) + ")");
				*valid = false;
			}
		}

		if(*valid) success++;
	}

	return success;
}                                       
//---------------------------------------------------------------------------
void __fastcall TStamperBase::ShowProperties(int AFileCount, int ADirCount,
	const TStamp& AAverage)
{
	TStringList* lines = new TStringList();

	// prepare contents
	lines->Add("[Contents]");
	lines->Add("File Count: " + FormatInt(AFileCount));
	lines->Add("Dir Count: " + FormatInt(ADirCount));
	lines->Add("File and Dir Total: " + FormatInt(AFileCount + ADirCount));

	// prepare average
	AnsiString line;
	FILETIME fileTime, localFileTime;
	SYSTEMTIME systemTime;
	AnsiString unknown = "(Unknown)";
	lines->Add("");
	lines->Add("[Average]");
	for(int i = 0; i < 3; i++) {
		switch(i) {
			case 0: fileTime = AAverage.CRTime; line = "CR Time: "; break;
			case 1: fileTime = AAverage.LWTime; line = "LW Time: "; break;
			case 2: fileTime = AAverage.LATime; line = "LA Time: "; break;
		}
		if(!IsValidFileTime(fileTime) ||
			FileTimeToLocalFileTime(&fileTime, &localFileTime) == FALSE ||
			FileTimeToSystemTime(&localFileTime, &systemTime) == FALSE) {
			line += unknown;
		}
		else {
			line += AnsiString().sprintf("%04d-%02d-%02d %02d:%02d:%02d.%03d",
				systemTime.wYear, systemTime.wMonth, systemTime.wDay, systemTime.wHour,
				systemTime.wMinute, systemTime.wSecond, systemTime.wMilliseconds);
		}
		lines->Add(line);
	}
	line = "UTC Bias: ";
	TIME_ZONE_INFORMATION timeZoneInformation;
	if(GetTimeZoneInformation(&timeZoneInformation) == 0xFFFFFFFF)
		line += unknown;
	else
		line += FormatInt(timeZoneInformation.Bias) + " minute(s)";
	lines->Add(line);

	// show dialog
	ShowMessageBox(lines->Text, MB_ICONINFORMATION);
	delete lines;
}
//---------------------------------------------------------------------------
TStamp __fastcall TStamperBase::StampsToAverage(const TStamps& AStamps) const
{
	const int size = 3;
	DWORD sums[][size] = {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}};
	int totals[] = {0, 0, 0};
	TStamp average;

	// calculate sums
	for(TStamps::const_iterator it = AStamps.begin(); it != AStamps.end() && !Terminated; it++) {
		if(CalculateSum(sums[0], size, it->CRTime)) totals[0]++;
		if(CalculateSum(sums[1], size, it->LWTime)) totals[1]++;
		if(CalculateSum(sums[2], size, it->LATime)) totals[2]++;
	}

	// calculate average
	CalculateAverage(average.CRTime, sums[0], size, totals[0]);
	CalculateAverage(average.LWTime, sums[1], size, totals[1]);
	CalculateAverage(average.LATime, sums[2], size, totals[2]);

	return average;
}
//---------------------------------------------------------------------------
int __fastcall TStamperBase::ShowMessageBox(const AnsiString& AText, int AFlags)
{
	FMessageBoxText = AText;
	FMessageBoxFlags = AFlags;
	Synchronize(ShowMessageBox_GetResult);
	return FMessageBoxResult;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TStamperBase::TStamperBase() : TThread(true)
{
	FreeOnTerminate = true;
}
//---------------------------------------------------------------------------
__fastcall TStamperBase::~TStamperBase()
{
	// do nothing
}
//---------------------------------------------------------------------------

