//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <values.h>  // contains BITSPERBYTE
#include "TStamperThread.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
static const int DWORD_BIT = BITSPERBYTE * sizeof(DWORD);
//---------------------------------------------------------------------------
// class TStamperThread : public TThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
AnsiString __fastcall TStamperThread::GetErrorLog() const
{
	return FErrorBuffer->Text;
}
//---------------------------------------------------------------------------
bool __fastcall TStamperThread::Get(TFileInfo& AFileInfo, TStringList* AErrorBuffer) const
{
	WIN32_FIND_DATA data;
	HANDLE handle = FindFirstFile(AFileInfo.FilePath.c_str(), &data);
	if(handle == INVALID_HANDLE_VALUE || handle == NULL) {
		if(AErrorBuffer != NULL)
			AErrorBuffer->Add(AnsiString("Failed to Get File Times: ") + AFileInfo.FilePath);
		return false;
	}
	FindClose(handle);

	AFileInfo.CRTime = data.ftCreationTime;
	AFileInfo.LWTime = data.ftLastWriteTime;
	AFileInfo.LATime = data.ftLastAccessTime;

	return true;
}
//---------------------------------------------------------------------------
bool __fastcall TStamperThread::Set(const TFileInfo& AFileInfo, TStringList* AErrorBuffer) const
{
	DWORD attributes;
	bool hasAttributes = false;
	if((AFileInfo.Flags & TFileInfo::FLAG_ANYATTR) == TFileInfo::FLAG_ANYATTR) {
		attributes = GetFileAttributes(AFileInfo.FilePath.c_str());
		hasAttributes = attributes != 0xFFFFFFFF &&
			SetFileAttributes(AFileInfo.FilePath.c_str(), FILE_ATTRIBUTE_NORMAL) != FALSE;
	}

	bool succeeded = false;

	HANDLE handle = CreateFile(AFileInfo.FilePath.c_str(), GENERIC_WRITE,
		FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL |
		(DirectoryExists(AFileInfo.FilePath) ? FILE_FLAG_BACKUP_SEMANTICS : 0), NULL);
	if(handle != INVALID_HANDLE_VALUE && handle != NULL) {
		if(SetFileTime(handle,
			(AFileInfo.Flags & TFileInfo::FLAG_CRTIME) == TFileInfo::FLAG_CRTIME ? &(AFileInfo.CRTime) : NULL,
			(AFileInfo.Flags & TFileInfo::FLAG_LATIME) == TFileInfo::FLAG_LATIME ? &(AFileInfo.LATime) : NULL,
			(AFileInfo.Flags & TFileInfo::FLAG_LWTIME) == TFileInfo::FLAG_LWTIME ? &(AFileInfo.LWTime) : NULL)
			!= FALSE) {
			succeeded = true;
		}
		CloseHandle(handle);
	}
	if(!succeeded) {
		if(AErrorBuffer != NULL)
			AErrorBuffer->Add(AnsiString("Failed to Set File Times: ") + AFileInfo.FilePath);
	}

	if(hasAttributes && SetFileAttributes(AFileInfo.FilePath.c_str(), attributes) == FALSE) {
		succeeded = false;
		if(AErrorBuffer != NULL) {
			AnsiString text;
			DWORD flags[] = {FILE_ATTRIBUTE_DIRECTORY, FILE_ATTRIBUTE_READONLY,
				FILE_ATTRIBUTE_HIDDEN, FILE_ATTRIBUTE_ARCHIVE,
				FILE_ATTRIBUTE_SYSTEM, FILE_ATTRIBUTE_TEMPORARY};
			for(int i = 0; i < 6; i++)
				text += (attributes & flags[i]) == flags[i] ? '+' : '-';
			AErrorBuffer->Add(AnsiString("Failed to Restore File Attrs: ") + AFileInfo.FilePath);
			AErrorBuffer->Add(AnsiString("Original File Attrs (DRHAST): ") + text);
		}
	}

	return succeeded;
}
//---------------------------------------------------------------------------
int __fastcall TStamperThread::CountDirs(const TFileInfos& AFileInfos) const
{
	int count = 0;

	for(TFileInfos::const_iterator it = AFileInfos.begin(); it != AFileInfos.end() && !Terminated; it++) {
		if(DirectoryExists(it->FilePath)) count++;
	}

	return count;
}
//---------------------------------------------------------------------------
void __fastcall TStamperThread::Average(TFileInfo& AAverage, const TFileInfos& AFileInfos) const
{
	const int size = 3;
	DWORD sums[][size] = {{0, 0, 0}, {0, 0, 0}, {0, 0, 0}};
	int totals[] = {0, 0, 0};

	// calculate sums
	for(TFileInfos::const_iterator it = AFileInfos.begin(); it != AFileInfos.end() && !Terminated; it++) {
		if(CalculateSum(sums[0], size, it->CRTime)) totals[0]++;
		if(CalculateSum(sums[1], size, it->LWTime)) totals[1]++;
		if(CalculateSum(sums[2], size, it->LATime)) totals[2]++;
	}

	// calculate average
	CalculateAverage(AAverage.CRTime, sums[0], size, totals[0]);
	CalculateAverage(AAverage.LWTime, sums[1], size, totals[1]);
	CalculateAverage(AAverage.LATime, sums[2], size, totals[2]);
}
//---------------------------------------------------------------------------
bool __fastcall TStamperThread::CalculateSum(DWORD* ASum, int ASize,
	const FILETIME& AFileTime) const
{
	if(!IsValidFileTime(AFileTime)) return false;

	CalculateSum(ASum, ASize, 0, AFileTime.dwLowDateTime);
	CalculateSum(ASum, ASize, 1, AFileTime.dwHighDateTime);

	return true;
}
//---------------------------------------------------------------------------
void __fastcall TStamperThread::CalculateSum(DWORD* ASum, int ASize,
	int APosition, DWORD AValue) const
{
	ASum[APosition] += AValue;
	if(ASum[APosition] < AValue && APosition < ASize - 1)
		CalculateSum(ASum, ASize, APosition + 1, 1);
}
//---------------------------------------------------------------------------
void __fastcall TStamperThread::CalculateAverage(FILETIME& AFileTime,
	DWORD* ASum, int ASize, int ATotal) const
{
	if(ATotal <= 0) {
		ClearFileTime(AFileTime);
		return;
	}

	int position;
	DWORDLONG n;
	for(int i = 0; i < ASize; i++) {
		position = ASize - 1 - i;
		n = ASum[position] + (i > 0 ? static_cast<DWORDLONG>(ASum[position + 1]) << DWORD_BIT : 0);
		switch(position) {
			case 0: AFileTime.dwLowDateTime = static_cast<DWORD>(n / ATotal); break;
			case 1: AFileTime.dwHighDateTime = static_cast<DWORD>(n / ATotal); break;
		}
		ASum[position] = static_cast<DWORD>(n % ATotal);
	}
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TStamperThread::Execute()
{
	// do nothing
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TStamperThread::TStamperThread() : TThread(true), FErrorBuffer(NULL)
{
	FreeOnTerminate = false;
	FErrorBuffer = new TStringList();
}
//---------------------------------------------------------------------------
__fastcall TStamperThread::~TStamperThread()
{
	delete FErrorBuffer;
}
//---------------------------------------------------------------------------
bool __fastcall TStamperThread::IsValidFileTime(const FILETIME& AFileTime)
{
	return !(AFileTime.dwLowDateTime == 0 && AFileTime.dwHighDateTime == 0) &&
		(static_cast<DWORDLONG>(AFileTime.dwHighDateTime) << DWORD_BIT) +
		AFileTime.dwLowDateTime < 0x8000000000000000;
}
//---------------------------------------------------------------------------
void __fastcall TStamperThread::ClearFileTime(FILETIME& AFileTime)
{
	AFileTime.dwLowDateTime = AFileTime.dwHighDateTime = 0;
}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
// class TFileGetThread : public TStamperThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TFileGetThread::Execute()
{
	try {
		FResults.Total = FFileInfos.size();

		// get FileInfos field
		for(unsigned int i = 0; i < FFileInfos.size() && !Terminated; i++) {
			if(!Get(FFileInfos[i], FErrorBuffer))
				FFileInfos.erase(FFileInfos.begin() + i--);
		}
		FFileGetData.FileInfos = FFileInfos;

		// get other fields
		TFileInfo average;
		Average(average, FFileInfos);
		FFileGetData.CRTime = average.CRTime;
		FFileGetData.LWTime = average.LWTime;
		FFileGetData.LATime = average.LATime;
		FFileGetData.DirCount = CountDirs(FFileInfos);

		FResults.Success = FFileInfos.size();
		FResults.Canceled = Terminated;
	}
	catch(...) {}
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TFileGetThread::TFileGetThread(const TFileInfos& AFileInfos) : TStamperThread()
{
	FFileInfos = AFileInfos;
}
//---------------------------------------------------------------------------
__fastcall TFileGetThread::~TFileGetThread()
{
	// do nothing
}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
// class TFileSetThread : public TStamperThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TFileSetThread::Execute()
{
	try {
		FResults.Total = FFileInfos.size();
		FResults.Success = 0;

		for(unsigned int i = 0; i < FFileInfos.size() && !Terminated; i++) {
			if(Set(FFileInfos[FForward ? i : (FFileInfos.size() - 1 - i)], FErrorBuffer))
				FResults.Success++;
		}

		FResults.Canceled = Terminated;
	}
	catch(...) {}
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TFileSetThread::TFileSetThread(const TFileInfos& AFileInfos, bool AForward) :
	TStamperThread(), FForward(AForward)
{
	FFileInfos = AFileInfos;
}
//---------------------------------------------------------------------------
__fastcall TFileSetThread::~TFileSetThread()
{
	// do nothin
}
//---------------------------------------------------------------------------

